"""
Line postprocessing
"""
from __future__ import annotations
from abc import ABC, abstractmethod


def linelimit(line: str, limit: int = 80) -> str:
    """
    Limit a line to a certain length by adding newlines after some commas.

    >>> linelimit("aa,bb,cc,dd", limit=6)
    'aa,bb,\\ncc,dd'
    >>> linelimit('abc=01,23,45,67,89,ab,cd,ef', limit=16)
    'abc=01,23,45,\\n  67,89,ab,cd,\\n  ef'
    >>> linelimit('aaaaaaaaaa=01,23,45,67,89,ab,cd,ef', limit=10)
    'aaaaaaaaaa=01,\\n  23,45,\\n  67,89,\\n  ab,cd,\\n  ef'

    Args:
        line: The line to limit.
        limit: The maximum length of the line.

    Returns:
        The line, limited to the given length.
    """
    if "," not in line:
        return line
    res = ""
    char_count = 0
    for chars in line.split(","):
        added_chars = chars + ","
        res += added_chars
        char_count += len(added_chars)
        if char_count + len(chars) > limit - 2:
            res += "\\\n  "
            char_count = 2
    if res.endswith(","):
        res = res[:-1]
    return res


class PostProcessor(ABC):
    """
    Abstract base class for post-processing.
    """
    @staticmethod
    @abstractmethod
    def postprocess(line: str) -> str:
        """
        Postprocess a line.

        Args:
            line: The line to postprocess.

        Returns:
            The postprocessed line.
        """


class PPNone(PostProcessor):
    """
    Post-processor that does nothing.
    """
    @staticmethod
    def postprocess(line: str) -> str:
        "Returns the line as is"
        return line


class PPLinelimit(PostProcessor):
    """
    Post-processor that limits a line to ~80 characters.
    """
    @staticmethod
    def postprocess(line: str) -> str:
        "Limits the line to ~80 characters"
        return linelimit(line)



"""
Parser for .reg files.
"""
# pylint: disable=invalid-name
from __future__ import annotations
import re
from dataclasses import dataclass
from pathlib import Path
from os import (
    PathLike,
    fspath,
)
from typing import (
    Any,
    ClassVar,
    Optional,
)

from .common import REG_ENCODING
from .types import (
    escape,
    RegPath,
    Value,
    value_from_str,
    Key,
)


@dataclass
class RegFileHeader:
    version: str
    has_bom: bool = False
    MIME: ClassVar[str] = "Windows Registry Editor Version {}"
    MIME_BOM: ClassVar[str] = "\ufeff"
    MIME_RE: ClassVar[str] = r"Windows Registry Editor Version (.+)"

    @classmethod
    def from_str(cls, data: str) -> "RegFileHeader":
        has_bom: bool = False
        mime = data.split("\n")[0]
        if mime[0] == cls.MIME_BOM:
            mime = mime[1:]
            has_bom = True
        match = re.match(cls.MIME_RE, mime)
        if not match:
            raise TypeError(f"""Not a registry file!
Unknown mime type: {escape(mime)}""")
        version = str(match.groups(0)[0])
        if version != "5.00":
            raise RuntimeError("Unsupported reg file version.")
        return cls(version, has_bom)

    def dump(self) -> str:
        res = self.MIME.format(self.version)
        if self.has_bom:
            res = self.MIME_BOM + res
        return res + "\n"


@dataclass
class RegFile:
    header: RegFileHeader
    keys: list[Key]
    root_key: Optional[RegPath]

    @classmethod
    def from_str(cls, data: str) -> "RegFile":
        lines: list[str] = data.replace("\\\n  ", "").split("\n")
        keys: list[Key] = []
        current_key: Optional[Key] = None
        header_str: str = ""
        header: Optional[RegFileHeader] = None
        root_key: Optional[Key] = None
        for line in lines:
            if line == "":
                continue
            if line.startswith("["):
                path = RegPath.from_str(line[1:-1])
                key = Key(path)
                if current_key is None:
                    header = RegFileHeader.from_str(header_str)
                current_key = key
                # The parent key always appears before its child
                if key.parent is None:
                    keys.append(key)
                    continue
                if not keys:
                    # Otherwhise it's a partial reg file
                    if not root_key:
                        root_key = key
                        continue
                    cpath = key.parent.path[len(root_key.path.path):]
                    if not cpath:
                        root_key.add_subkey(key)
                        continue
                root_keys = keys.copy()
                if root_key:
                    root_keys += [root_key]
                parent_found = False
                for sub_root_key in root_keys:
                    srk_path = sub_root_key.path
                    if not key.parent.startswith(srk_path):
                        continue
                    kp_path = key.parent.path
                    for _ in srk_path:
                        kp_path.pop(0)
                    if len(kp_path) == 0:
                        sub_root_key.add_subkey(key)
                        parent_found = True
                        break
                    parent_key = sub_root_key.find_key(kp_path)
                    parent_key.add_subkey(key)
                    parent_found = True
                    break
                if parent_found:
                    continue
                raise ValueError(f"{escape(key.name)}'s parent not found'")
            if current_key is None:
                header_str += line + "\n"
                continue
            value: Value[Any] = value_from_str(line)
            current_key.add_value(value)
        if header is None:
            raise ValueError("Header not found")
        # Add missing keys
        root_path = None
        if root_key:
            tmp = root_key
            tmp_parent: Key
            while tmp.parent:
                tmp_parent = Key(tmp.parent)
                tmp_parent.add_subkey(tmp)
                tmp = tmp_parent
            keys.append(tmp)
            root_path = root_key.path
        return cls(header, keys, root_path)

    @classmethod
    def from_path(cls, path: PathLike):
        path = Path(fspath(path))
        return cls.from_str(path.read_text(REG_ENCODING))

    def __getitem__(self, name: str) -> Key:
        for key in self.keys:
            if key.name == name:
                return key
        raise KeyError(name)

    def find_key(self, path: list[str] | str) -> Optional[Key]:
        """
        Finds key by path.
        """
        if isinstance(path, str):
            path = path.split("\\")
        search = path.pop(0)
        for key in self.keys:
            if key.name == search:
                if len(path) == 0:
                    return key
                return key.find_key(path)
        return None

    @property
    def HKEY_CLASSES_ROOT(self) -> Key:
        return self["HKEY_CLASSES_ROOT"]

    @property
    def HKEY_CURRENT_USER(self) -> Key:
        return self["HKEY_CURRENT_USER"]

    @property
    def HKEY_LOCAL_MACHINE(self) -> Key:
        return self["HKEY_LOCAL_MACHINE"]

    @property
    def HKEY_USERS(self) -> Key:
        return self["HKEY_USERS"]

    @property
    def HKEY_CURRENT_CONFIG(self) -> Key:
        return self["HKEY_CURRENT_CONFIG"]

    def dump(self) -> str:
        """
        Dump registry file to string.
        Remember that Windows's reg files use CRLF as line endings.

        Returns:
            str: Registry file as string.
        """

        res = self.header.dump() + "\n"
        if self.root_key is not None:
            root = self.find_key(str(self.root_key))
            assert root is not None
            keys = [root]
        else:
            keys = self.keys
        all_keys = []
        while keys:
            key = keys.pop(0)
            all_keys.append(key)
            keys = key.subkeys + keys
        while len(all_keys) > 0:
            key = all_keys.pop(0)
            res += key.dump()
        return res



"""
Type detection for reg types

Yes I know what [MIME](https://en.wikipedia.org/wiki/MIME) means.
"""
from __future__ import annotations
from abc import ABC, abstractmethod

from .common import escape


class MIME(ABC):
    """
    Abstract base class for type detection.
    """
    mime: str

    @classmethod
    @abstractmethod
    def is_type_of(cls, value: str) -> bool:
        "Check if value is of this type."

    @classmethod
    @abstractmethod
    def mime_remove(cls, value: str) -> str:
        "Remove characteristics associated with the type"

    @classmethod
    @abstractmethod
    def mime_add(cls, value: str) -> str:
        "Re-add characteristics associated with the type"


class MIMEExact(MIME):
    """
    Tries to exactly match a string before the colon.
    """
    @classmethod
    def is_type_of(cls, value: str) -> bool:
        return value.split(":")[0] == cls.mime

    @classmethod
    def mime_remove(cls, value: str) -> str:
        return value[len(cls.mime) + 1:]

    @classmethod
    def mime_add(cls, value: str) -> str:
        return f"{cls.mime}:{value}"


class MIMEQuoted(MIME):
    """
    Checks if the string is quoted.
    """
    @classmethod
    def is_type_of(cls, value: str) -> bool:
        return (value.startswith(cls.mime) and value.endswith(cls.mime))

    @classmethod
    def mime_remove(cls, value: str) -> str:
        return value[len(cls.mime):-len(cls.mime)]

    @classmethod
    def mime_add(cls, value: str) -> str:
        return escape(value)



"""
Regedit registry file parser.

This module provides a parser for the windows registry file format.
"""
from .common import (
    REG_ENCODING,
)
from .mime import (
    MIME,
    MIMEExact,
    MIMEQuoted
)
from .postprocessing import (
    linelimit,
    PostProcessor,
    PPNone,
    PPLinelimit,
)
from .types import (
    RegPath,
    Value,
    REG_DWORD,
    REG_QWORD,
    REG_SZ,
    REG_BINARY,
    REG_EXPAND_SZ,
    REG_MULTI_SZ,
    mimemap,
    value_from_str,
    Key,
)
from .main import (
    RegFile,
)

__all__ = [
    "REG_ENCODING",
    "MIME",
    "MIMEExact",
    "MIMEQuoted",
    "linelimit",
    "PostProcessor",
    "PPNone",
    "PPLinelimit",
    "RegPath",
    "Value",
    "REG_DWORD",
    "REG_QWORD",
    "REG_SZ",
    "REG_BINARY",
    "REG_EXPAND_SZ",
    "REG_MULTI_SZ",
    "mimemap",
    "value_from_str",
    "Key",
    "RegFile",
]



"""
Types of data in registry files.

Data type descriptions are copied from
https://learn.microsoft.com/en-us/windows/win32/sysinfo/registry-value-types
"""
# pylint: disable=invalid-name
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from textwrap import wrap
from typing import (
    Generic,
    Iterable,
    Optional,
    Type,
    TypeVar,
)
from typing_extensions import (
    SupportsIndex,
)

from .common import (
    REG_ENCODING,
    escape,
)
from .mime import (
    MIMEExact,
    MIMEQuoted,
)
from .postprocessing import (
    PPNone,
    PPLinelimit,
)


@dataclass
class RegPath(Iterable):
    """
    A path to a registry key.
    """
    path: list[str]

    @classmethod
    def from_str(cls, path: str):
        """
        Create a RegPath from a string.

        Args:
            path: The path to parse.

        Returns:
            A RegPath.
        """
        return cls(path.split("\\"))

    @property
    def parent(self) -> Optional["RegPath"]:
        """
        Get the parent path.
        """
        if len(self.path) == 1:
            return None
        return RegPath(self.path[:-1])

    def startswith(
        self,
        __prefix: "RegPath",
        __start: SupportsIndex | None = None,
        __end: SupportsIndex | None = None
    ) -> bool:
        return str(self).startswith(str(__prefix), __start, __end)

    def __getitem__(self, index: int) -> str:
        return self.path[index]

    def __iter__(self):
        return iter(self.path)

    def __str__(self) -> str:
        return "\\".join(self.path)

    def __len__(self) -> int:
        return self.path.__len__()


T = TypeVar("T")


@dataclass
class Value(MIMEExact, PPNone, ABC, Generic[T]):
    """
    Generic value of a registry key.
    """
    name: str
    value: T
    name_quoted: bool = False

    @property
    def typeof(self) -> str:
        return self.__class__.__name__

    @classmethod
    def from_str(cls, name: str, value: str):
        """
        Create a Value from a string.

        Args:
            name: The name of the value.
            value: The value of the value.

        Returns:
            A Value.
        """
        name_quoted = False
        if name.startswith("\""):
            name = name[1:-1]
            name_quoted = True
        value = cls.mime_remove(value)
        return cls(name, cls.val_conv(value), name_quoted)

    @classmethod
    @abstractmethod
    def val_conv(cls, value: str) -> T:
        """
        Convert a string to the appropriate type.

        Args:
            value: The string to convert.

        Returns:
            The converted value.
        """

    @classmethod
    @abstractmethod
    def val_deconv(cls, value: T) -> str:
        """
        Convert a value to a string.

        Args:
            value: The value to convert.

        Returns:
            The converted string.
        """

    def to_str(self) -> str:
        """
        Convert the value to a string.

        Returns:
            The converted string
        """
        name = self.name
        if self.name_quoted:
            name = escape(name)
        value = self.val_deconv(self.value)
        value = self.mime_add(value)
        return self.postprocess(f"{name}={value}")

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.name}: {self.typeof} = {self.value}"


class REG_DWORD(Value):
    "A 32-bit number."
    value: int
    mime: str = "dword"

    @classmethod
    def val_conv(cls, value: str) -> int:
        return int(value, base=16)

    @classmethod
    def val_deconv(cls, value: int) -> str:
        return hex(value)[2:].zfill(8)


class REG_QWORD(Value):
    "A 64-bit number."
    value: int
    mime: str = "hex(b)"

    @classmethod
    def val_conv(cls, value: str) -> int:
        ordfix = value.split(",")
        ordfix.reverse()
        fixedval = "".join(ordfix)
        return int(fixedval, base=16)

    @classmethod
    def val_deconv(cls, value: int) -> str:
        vstr = hex(value)[2:].zfill(16)
        ordfix = wrap(vstr, 2)
        ordfix.reverse()
        return ",".join(ordfix)


class REG_SZ(MIMEQuoted, Value):
    """
    A null-terminated string.
    It's either a Unicode or an ANSI string, depending on whether you use the
    Unicode or ANSI functions.
    """
    value: str
    mime: str = "\""

    @classmethod
    def val_conv(cls, value: str) -> str:
        return value

    @classmethod
    def val_deconv(cls, value: str) -> str:
        return value


class REG_BINARY(PPLinelimit, Value):
    "Binary data in any form."
    value: bytes
    mime: str = "hex"

    @classmethod
    def val_conv(cls, value: str) -> bytes:
        vsplit = value.replace(",", "")
        return bytes.fromhex(vsplit)

    @classmethod
    def val_deconv(cls, value: bytes) -> str:
        vsplit = wrap(value.hex(), 2)
        return ",".join(vsplit)


class REG_EXPAND_SZ(PPLinelimit, Value):
    """
    A null-terminated string that contains unexpanded references to environment
    variables, for example, %PATH%. It's either a Unicode or an ANSI string,
    depending on whether you use the Unicode or ANSI functions. To expand the
    environment variable references, use the
    [ExpandEnvironmentStrings](https://learn.microsoft.com/en-us/windows/win32/api/processenv/nf-processenv-expandenvironmentstringsa)
    function.
    """
    value: str
    mime: str = "hex(2)"

    @classmethod
    def val_conv(cls, value: str) -> str:
        vbytes = REG_BINARY.val_conv(value)
        return str(vbytes, encoding=REG_ENCODING)

    @classmethod
    def val_deconv(cls, value: str) -> str:
        vbytes = bytes(value, encoding=REG_ENCODING)
        return REG_BINARY.val_deconv(vbytes)


class REG_MULTI_SZ(PPLinelimit, Value):
    """
    A sequence of null-terminated strings, terminated by an empty string (\\0).
    The following is an example:
    String1\\0String2\\0String3\\0LastString\\0\\0.
    The first \\0 terminates the first string,
    the second-from-last \\0 terminates the last string,
    and the final \\0 terminates the sequence.
    Note that the final terminator must be factored into the length of the string.
    """
    value: list[str]
    mime: str = "hex(7)"

    @classmethod
    def val_conv(cls, value: str) -> list[str]:
        res = REG_EXPAND_SZ.val_conv(value).split("\x00")
        for _ in range(2):
            res.pop(-1)
        return res

    @classmethod
    def val_deconv(cls, value: list[str]) -> str:
        for _ in range(2):
            value.append("")
        return REG_EXPAND_SZ.val_deconv("\x00".join(value))


# NOTE: Not implemented
class REG_RESOURCE_REQUIREMENTS_LIST(Value):
    value: list[str]
    mime: str = "hex(a)"

    @classmethod
    def is_type_of(cls, value: str) -> bool:
        return False

    @classmethod
    def val_conv(cls, value: str) -> list[str]:
        return []

    @classmethod
    def val_deconv(cls, value: list[str]) -> str:
        return ""


# TODO: Remove and add support for whatever this is:
# hex(ffff0011):ff
class REG_SKIP(Value):
    value: str
    mime: str = "hex(ff"

    @classmethod
    def is_type_of(cls, value: str) -> bool:
        return False

    @classmethod
    def val_conv(cls, value: str) -> str:
        return ""

    @classmethod
    def val_deconv(cls, value: str) -> str:
        return ""


mimemap: list[type[Value]] = [
    REG_DWORD,
    REG_QWORD,
    REG_EXPAND_SZ,
    REG_MULTI_SZ,
    REG_BINARY,
    REG_SZ,
]


def value_from_str(text: str) -> Value:
    """
    Returns a Value object from a string.

    Raises:
        TypeError: If the string value type could not be determined.
    """
    name: str = text.split("=")[0]
    value: str = text.split("=")[1]
    reg_type: Type[Value]
    for reg_type in mimemap:
        if not reg_type.is_type_of(value):
            continue
        res = reg_type.from_str(name, value)
        assert res is not None  # NOTE: This is a hack
        return res
    raise TypeError(f"Unknown type for line:\n{escape(text)}")


@dataclass
class Key:
    path: RegPath
    subkeys: list["Key"] = field(default_factory=list)
    values: list[Value] = field(default_factory=list)

    @property
    def name(self) -> str:
        return self.path[-1]

    @property
    def parent(self) -> Optional[RegPath]:
        return self.path.parent

    def add_subkey(self, key: "Key"):
        self.subkeys.append(key)

    def __getitem__(self, name: str) -> "Key":
        for key in self.subkeys:
            if key.name == name:
                return key
        raise KeyError(name)

    def add_value(self, value: Value):
        self.values.append(value)

    def get_value(self, name: str) -> Optional[Value]:
        for value in self.values:
            if value.name == name:
                return value
        return None

    def find_key(self, path: list[str]) -> "Key":
        if len(path) < 1:
            raise ValueError()
        search = path.pop(0)
        key = self[search]
        if len(path) == 0:
            return key
        return key.find_key(path)

    def __str__(self):
        return self.path.__str__()

    def __repr__(self):
        return f"Key({self.path})"

    def dump(self) -> str:
        res = f"[{self.path}]\n"
        for value in self.values:
            res += value.to_str() + "\n"
        res += "\n"
        return res
